///////////////////////////////////////////////////////////////////////////////
//
// Redefinition of debugging include
//
//  Debugging include type switching functionality. This enables the 
//  feature to include this file anywhere in the code and switch 
//  debugging mode on or off as desired.
//
//  Example:
//
//      [... some code ...]
//
//      on global scope:
//
//      // Turn on debugging output
//      #define LIB_DEBUG
//      #include "lib_debug.mqh"
//
//      void some_function()
//      { ... }
//
//      // Turn debugging off again
//      #undef LIB_DEBUG
//      #include "lib_debug.mqh"
//
//
//

// Switch debug mode
#ifdef LIB_DEBUG_MQH_INCLUDED
#ifdef LIB_DEBUG_MQH_RUNTIME_MODE
    #ifdef LIB_DEBUG
    #undef LIB_DEBUG_MQH_INCLUDED
    #endif
#else 
    #ifdef LIB_DEBUG_MQH_DEBUG_MODE
        #ifndef LIB_DEBUG
        #undef LIB_DEBUG_MQH_INCLUDED
        #endif
    #endif
#endif
#endif

// Include file flag
#ifndef LIB_DEBUG_MQH_INCLUDED
#define LIB_DEBUG_MQH_INCLUDED

//
///////////////////////////////////////////////////////////////////////////////



/**********************************************************************************
 * Copyright (C) 2010-2020 Dominik Egert <dominik.egert@freie-netze.de>
 *
 * This file holds debug support code.
 *
 * This file may be copied and/or distributed at free will.
 * Author Dominik Egert / Freie Netze UG.
 **********************************************************************************
 *
 *  File information
 *  ================
 *
 *  Version: 2.00-rc1
 *
 *
 *  Usage:
 *
 *      Defines macros for supporting debugging code.
 *
 *      Macros beginning with DBG_STR_* will represent a string.
 *      Macros beginning with DBG_MSG_* will printf information.
 *
 *      Following output and string macros are defined:
 *
 *          DBG_STR()                       // Trace info prefix.
 *          DBG_STR_PERSIST()               // Trace info prefix, message string will persist into release build.
 *          DBG_STR_VAR()                   // Convert variable value to string
 *
 *          DBG_MSG()                       // Will print out a text prefixed with trace information
 *          DBG_MSG_VAR()                   // Print a variable
 *
 *          DBG_MSG_TRACE_BEGIN             // Output >Trace begin<
 *          DBG_MSG_TRACE_END               // Output >Trace end<
 *          DBG_MSG_TRACE_RETURN            // Replace return(void);
 *          DBG_MSG_TRACE_RETURN_VAR(x)     // Replace return(x);
 *
 *          DBG_STR_COMMENT()               // A comment string only valid in debug, removed in release
 *
 *          DBG_STR_OBJ_RTTI                // Resolve object name to string
 *          DBG_MSG_OBJ_RTTI                // Print object name
 *
 *
 *      Vardump for arrays:
 *
 *          DBG_MSG_VARDUMP(x)              // Dump a variables or arrays content
 *
 *
 *      To support assert debugging code, following macros are defined:
 *
 *          DBG_ASSERT(condition, message)                      // Default assert macro. Will end execution by crash code.
 *          DBG_ASSERT_LOG(condition, message)                  // Logging only assert macro. Execution will continue normally.
 *          DBG_ASSERT_RETURN_VOID(condition, message)          // Return assert macro. Will set a function return if condition is met.
 *          DBG_ASSERT_RETURN(condition, message, return_value) // Return assert macro. Will set a function return if condition is met and return given value.
 *
 *
 *      Software Break points:
 *
 *          DBG_SLEEP_SECONDS(seconds)          // Will insert a realtime sleep into execution
 *          DBG_SOFT_BREAKPOINT                 // Will halt execution for given timeout
 *          DBG_SOFT_BREAKPOINT_TS(timestamp)   // Will halt for timeout seconds at (TimeCurrent() - (TimeCurrent() % PeriodSeconds()) == timestamp)
 *          DBG_SOFT_BREAKPOINT_CONDITION(x)    // Will halt execution if condition is met
 *          DBG_SOFT_BREAKPOINT_EXEC_TIME(x)    // Will halt execution after given execution time runtime
 *
 *
 *      Loop tracing:
 *
 *          DBG_TRACE_LOOP_BEGIN                // A new loop will be defined
 *          DBG_TRACE_LOOP_START                // At the beginning of a loop
 *          DBG_TRACE_LOOP_FINISH               // At the end of a loop
 *          DBG_TRACE_LOOP_END                  // After a loop is done, will print out stats
 *
 *          DBG_TRACE_LOOP_BEGIN_ID(id)         // A new loop counter with id will be defined
 *          DBG_TRACE_LOOP_START_ID(id)         // Start loop counter with ID at the beginning of a loop
 *          DBG_TRACE_LOOP_FINISH_ID(id)        // Finish loop counter with ID at the end of a loop
 *          DBG_TRACE_LOOP_END_ID(id)           // After a loop is done, will print out stats for this ID
 *
 *
 *  Hint:
 *
 *      All debug macros will be removed when build as runtime.
 *
 *
 *  Example:
 *
 *      Function definition for call tracing:
 *
 *          Define two macros for replacement to be able to turn on/off tracing
 *          for a specific function. This is done here.
 *          To enable tracing a define will switch trace code on.
 *
 *          #define DBG_TRACE_SOME_FUNCTION
 *
 *          Later in the function body, the macro xxx_RETURN will be used instead
 *          of the usual "return" command. This must be adjusted to the functions return value
 *          and the appropiate macro needs to be selected below.
 *
 *          /////////////////////////////////////
 *          // Function debug trace code
 *
 *          #ifdef DBG_TRACE_SOME_FUNCTION
 *              #undef DBG_TRACE_SOME_FUNCTION
 *              #define DBG_TRACE_SOME_FUNCTION(x) x
 *              #define DBG_TRACE_SOME_FUNCTION_RETURN(x) DBG_MSG_TRACE_RETURN_VAR(x)
 *          #else
 *              #define DBG_TRACE_SOME_FUNCTION(x)
 *              #define DBG_TRACE_SOME_FUNCTION_RETURN(x) DBG_MSG_NOTRACE_RETURN_VAR(x)
 *          #endif
 *
 *          /////////////////////////////////////
 *
 *          Return value example:
 *
 *              #define xxx_RETURN(x) DBG_MSG_TRACE_RETURN_VAR(x)
 *
 *              Here the return value will be integer. it is important to match
 *              the return value to its actual type.
 *
 *              When using Performance Counters, the Return Macro gets extended
 *              with the macro PERF_COUNTER_END in RELEASE. This will print out
 *              the statistics of the functions performance counter value.
 *
 *
 *          double some_function(const int index = NULL)
 *          {
 *              DBG_TRACE_SOME_FUNCTION(DBG_MSG_TRACE_BEGIN);
 *              DBG_TRACE_SOME_FUNCTION(DBG_MSG_INT(index));
 *              PERF_COUNTER_BEGIN;
 *
 *              DBG_TRACE_SOME_FUNCTION(DBG_TRACE_LOOP_BEGIN);
 *              for(int cnt = (int)some_periods; ((cnt > NULL) && (!_StopFlag)); cnt--)
 *              {
 *                  DBG_TRACE_SOME_FUNCTION(DBG_TRACE_LOOP_START);
 *
 *                  // Basic calculations
 *                  a += a * cnt;
 *
 *                  // Some check
 *                  if(a)
 *                  { continue; }
 *
 *                  DBG_TRACE_SOME_FUNCTION(DBG_TRACE_LOOP_FINISH);
 *              }
 *              DBG_TRACE_SOME_FUNCTION(DBG_TRACE_LOOP_END);
 *
 *              // Assert value is greater than NULL
 *              DBG_ASSERT((some_testing_value > NULL), DBG_MSG_VAR(some_testing_value));
 *
 *              // Return
 *              DBG_TRACE_SOME_FUNCTION_RETURN(some_double_value);
 *          }
 *
 *
 *      Notice:
 *
 *          This line of code is the function return command.
 *          When passing a NULL value to a makro, you need to
 *          cast the NULL-Makro according to the return value type
 *          like this:
 *
 *              // Return
 *              DBG_TRACE_SOME_FUNCTION_RETURN((double)NULL);
 *
 *
 *  Performance tracing
 *
 *      These macros are used to have performance indications on
 *      the execution speed of a function or block of code.
 *
 *      The macros without parameter are specificated on the function
 *      itself, if more than one counter within the same function or
 *      execution block is needed, they can be created by using the
 *      macros with the parameter. Pass any type of ID to it and use
 *      it later to close this counter again.
 *
 *      Performance macros will only be available in RELEASE versions.
 *      They are stripped of in debugging as well as if disabled.
 *
 *      Defined performance macros
 *
 *          PERF_COUNTER_BEGIN                  // Begin measurement of execution time
 *          PERF_COUNTER_END                    // End performance measurement and show stats
 *
 *          PERF_COUNTER_DEFINE_ID(x)           // Create function global performance counter
 *          PERF_COUNTER_SET_ID(x)              // Begin measurement of execution time
 *          PERF_COUNTER_CLOSE_ID(x)            // End performance measurement and show stats
 *
 *
 *  EX5 Program info
 *
 *      A special macro is defined to deliver program information and can be
 *      displyed ie at program start. (Very helpful to determin if the program is
 *      running in debug or release mode)
 *
 *              DBG_STR_EX5_FILEINFO            // EX5 program information
 *
 *
 *  Available configuration options:
 *
 *      Basic settings are expposed below. Here is a full list of options:
 *
 *          Function options:
 *
 *              LIB_DEBUG                       // Force enable debugging support
 *              LIB_PERF_PROFILING              // Enable performance counters (Only in release mode)
 *              DBG_VARDUMP_PAUSE               // Vardump pause before continuing execution
 *              DBG_SOFT_BKP_TIMEOUT            // Breakpoint timeout before continuing execution
 *              DBG_TRACE_EXEC_DELAY_MS         // When tracing function calls, this will insert a delay in execution
 *              DBG_CRASH_CODE                  // Code used to crash the program if an assert() fails
 *
 *          Output format options:
 *
 *              DBG_ASSERT_MSG_TXT              // Prefix to assert messages
 *              DBG_TRACE_BEGIN_MSG_TXT         // Trace function call begin prefix
 *              DBG_TRACE_END_MSG_TXT           // Trace function call end prefix
 *              DBG_OUTPUT_PREFIX               // General debug output prefix
 *              DBG_OUTPUT_FORMAT               // Format string applied to debug output
 *              DBG_DOUBLE_OUTPUT_DIGITS        // Digits debugger wil loutput on double values
 *
 *      If not defined, the default applies automatically.
 *
 *          Following list of defaults is configured as preset:
 *
 *              DBG_SOFT_BKP_TIMEOUT            120                                         // Wait 2 minutes
 *              DBG_TRACE_EXEC_DELAY_MS         0x00                                        // Disabled by default
 *              DBG_CRASH_CODE                  { int i = 1; i = (i / (i - i)); }           // Zero divide abnormal program termination
 *
 *              DBG_ASSERT_MSG_TXT              "### --- Assertion failed!"
 *              DBG_TRACE_BEGIN_MSG_TXT         ">>>Function begin<<<"
 *              DBG_TRACE_END_MSG_TXT           ">>>Function end<<<"
 *              DBG_OUTPUT_PREFIX               "   "                                       // Four white spaces
 *              DBG_OUTPUT_STRING               "%s%s, %s, line: %i%s"
 *              DBG_OUTPUT_FORMAT               DBG_OUTPUT_STRING, ((prefix == NULL) ? NULL : prefix + ", "), __FILE__, __FUNCSIG__, __LINE__, ((message == NULL) ? NULL : ", " + message)
 *              DBG_DOUBLE_OUTPUT_DIGITS        SymbolInfoInteger(Symbol(), SYMBOL_DIGITS)
 *
*/



//////////////////////////////////////////
//
// Debugger switches and definitions
//

// Force enable debugger
//#define LIB_DEBUG

// Enable performance tracer
//#define LIB_PERF_PROFILING

// Pause execution after a vardump calls
#define DBG_VARDUMP_PAUSE 0

// Soft breakpoint timeout in seconds
//#define DBG_SOFT_BKP_TIMEOUT 120

// Execution delay inserted into trace runs in milliseconds
//#define DBG_TRACE_EXEC_DELAY_MS 25

// Crash code definition
//#define DBG_CRASH_CODE { int i = 1; i = (i / (i - i)); }

//
//////////////////////////////////////////



//////////////////////////////////////////
//
// Debugger output format setup
//

// Set default assert message
//#define DBG_ASSERT_MSG_TXT "### --- Assertion failed!"

// Set default trace begin messages
//#define DBG_TRACE_BEGIN_MSG_TXT ">>>Function begin<<<"

// Set default trace end messages
//#define DBG_TRACE_END_MSG_TXT ">>>Function end<<<"

// Default output prefix
//#define DBG_OUTPUT_PREFIX "   "

// Default file location string
//#define DBG_CODE_LOCATION_STRING __FILE__, __FUNCTION__, __LINE__

// Output format definition
//#define DBG_OUTPUT_STRING "%s>%s< %s(){ @%i: %s }"
//#define DBG_OUTPUT_FORMAT(prefix, message)  DBG_OUTPUT_STRING, ((prefix == "") ? "" : prefix + " "), DBG_CODE_LOCATION_STRING, message

// Float and double output precision
//#define DBG_FLOAT_OUTPUT_DIGITS SymbolInfoInteger(Symbol(), SYMBOL_DIGITS)
//#define DBG_DOUBLE_OUTPUT_DIGITS SymbolInfoInteger(Symbol(), SYMBOL_DIGITS)


//
//////////////////////////////////////////









//*********************************************************************************************************************************************************/
// Debugging support
//


/////////////////////////////////////////////////////////////////////////////////////////////////////
//
// Auto enable/disable by environment
//

// Autoenable
#ifdef _DEBUG
    #ifndef LIB_DEBUG
    #define LIB_DEBUG
    #endif
    #define DBG_DEBUGGER_FLAG "_DEBUG defined"
#else
    #ifdef MQL_DEBUG
        #ifndef LIB_DEBUG
        #define LIB_DEBUG
        #endif
        #define DBG_DEBUGGER_FLAG "MQL_DEBUG defined"
    #endif
#endif

// Environment check
#ifndef __MQL5__
#ifdef LIB_DEBUG
#undef LIB_DEBUG
#endif
#endif

//
/////////////////////////////////////////////////////////////////////////////////////////////////////





#ifdef LIB_DEBUG
/////////////////////////////////////////////////////////////////////////////////////////////////////
//
// Debugger mode
//

// Set debugging mode
#define LIB_DEBUG_MQH_DEBUG_MODE

// Set _DEBUG flag if not set
#ifndef _DEBUG
#define _DEBUG false
#endif

// Clear any existing definitions
#ifdef DBG_CODE_LOCATION
#undef DBG_CODE_LOCATION
#endif
#ifdef DBG_DEBUGGER_FLAG_STATE
#undef DBG_DEBUGGER_FLAG_STATE
#endif
#ifdef DBG_STR_EX5_FILEINFO
#undef DBG_STR_EX5_FILEINFO
#endif
#ifdef DBG_STR
#undef DBG_STR
#endif
#ifdef DBG_STR_PERSIST
#undef DBG_STR_PERSIST
#endif
#ifdef DBG_STR_VAR
#undef DBG_STR_VAR
#endif
#ifdef DBG_MSG
#undef DBG_MSG
#endif
#ifdef DBG_MSG_PERSIST
#undef DBG_MSG_PERSIST
#endif
#ifdef DBG_MSG_VAR
#undef DBG_MSG_VAR
#endif
#ifdef DBG_MSG_VARDUMP
#undef DBG_MSG_VARDUMP
#endif
#ifdef DBG_MSG_LISTDUMP
#undef DBG_MSG_LISTDUMP
#endif
#ifdef DBG_MSG_TRACE_BEGIN
#undef DBG_MSG_TRACE_BEGIN
#endif
#ifdef DBG_MSG_TRACE_END
#undef DBG_MSG_TRACE_END
#endif
#ifdef DBG_MSG_TRACE_RETURN
#undef DBG_MSG_TRACE_RETURN
#endif
#ifdef DBG_MSG_TRACE_RETURN_VAR
#undef DBG_MSG_TRACE_RETURN_VAR
#endif
#ifdef DBG_MSG_NOTRACE_RETURN
#undef DBG_MSG_NOTRACE_RETURN
#endif
#ifdef DBG_MSG_NOTRACE_RETURN_VAR
#undef DBG_MSG_NOTRACE_RETURN_VAR
#endif
#ifdef DBG_MSG_UNINIT_RESOLVER
#undef DBG_MSG_UNINIT_RESOLVER
#endif
#ifdef DBG_STR_COMMENT
#undef DBG_STR_COMMENT
#endif
#ifdef DBG_ASSERT
#undef DBG_ASSERT
#endif
#ifdef DBG_ASSERT_LOG
#undef DBG_ASSERT_LOG
#endif
#ifdef DBG_ASSERT_RETURN
#undef DBG_ASSERT_RETURN
#endif
#ifdef DBG_ASSERT_RETURN_VAR
#undef DBG_ASSERT_RETURN_VAR
#endif
#ifdef DBG_SLEEP_SECONDS
#undef DBG_SLEEP_SECONDS
#endif
#ifdef DBG_SOFT_BREAKPOINT
#undef DBG_SOFT_BREAKPOINT
#endif
#ifdef DBG_SOFT_BREAKPOINT_TS
#undef DBG_SOFT_BREAKPOINT_TS
#endif
#ifdef DBG_SOFT_BREAKPOINT_CONDITION
#undef DBG_SOFT_BREAKPOINT_CONDITION
#endif
#ifdef DBG_SOFT_BREAKPOINT_EXEC_TIME
#undef DBG_SOFT_BREAKPOINT_EXEC_TIME
#endif
#ifdef DBG_TRACE_LOOP_BEGIN
#undef DBG_TRACE_LOOP_BEGIN
#endif
#ifdef DBG_TRACE_LOOP_START
#undef DBG_TRACE_LOOP_START
#endif
#ifdef DBG_TRACE_LOOP_FINISH
#undef DBG_TRACE_LOOP_FINISH
#endif
#ifdef DBG_TRACE_LOOP_END
#undef DBG_TRACE_LOOP_END
#endif
#ifdef DBG_TRACE_LOOP_BEGIN_ID
#undef DBG_TRACE_LOOP_BEGIN_ID
#endif
#ifdef DBG_TRACE_LOOP_START_ID
#undef DBG_TRACE_LOOP_START_ID
#endif
#ifdef DBG_TRACE_LOOP_FINISH_ID
#undef DBG_TRACE_LOOP_FINISH_ID
#endif
#ifdef DBG_TRACE_LOOP_END_ID
#undef DBG_TRACE_LOOP_END_ID
#endif

//
/////////////////////////////////////////////////////////////////////////////////////////////////////





/////////////////////////////////////////////////////////////////////////////////////////////////////
//
// Debugger default configuration
//

// Code location string
#ifndef DBG_CODE_LOCATION_STRING
#define DBG_CODE_LOCATION_STRING __FILE__, __FUNCTION__, __LINE__
#endif
#define DBG_CODE_LOCATION(x) StringFormat("%s %s %s %i", x, DBG_CODE_LOCATION_STRING)

// Output format definition
#ifndef DBG_OUTPUT_FORMAT
#define DBG_OUTPUT_STRING "%s>%s< %s(){ @%i: %s }"
#define DBG_OUTPUT_FORMAT(prefix, message)  DBG_OUTPUT_STRING, ((prefix == "") ? "" : prefix + " "), DBG_CODE_LOCATION_STRING, message
#endif

// Float output precision
#ifndef DBG_FLOAT_OUTPUT_DIGITS
#define DBG_FLOAT_OUTPUT_DIGITS 9
#endif

// Double output precision
#ifndef DBG_DOUBLE_OUTPUT_DIGITS
#define DBG_DOUBLE_OUTPUT_DIGITS 16
#endif

// Crash code definition
#ifndef DBG_CRASH_CODE
#define DBG_CRASH_CODE { int i = 1; i = (i / (i - i)); }
#endif

// Set default assert message
#ifndef DBG_ASSERT_MSG_TXT
#define DBG_ASSERT_MSG_TXT "### --- Assertion failed!"
#endif

// Set default trace begin messages
#ifndef DBG_TRACE_BEGIN_MSG_TXT
#define DBG_TRACE_BEGIN_MSG_TXT ">>>Function begin<<<"
#endif

// Set default trace end messages
#ifndef DBG_TRACE_END_MSG_TXT
#define DBG_TRACE_END_MSG_TXT ">>>Function end<<<"
#endif

// Default output prefix
#ifndef DBG_OUTPUT_PREFIX
#define DBG_OUTPUT_PREFIX "  "
#endif

// Overwrite mode
#ifndef DBG_DEBUGGER_FLAG
#define DBG_DEBUGGER_FLAG "LIB_DEBUG overwrite mode"
#define DBG_DEBUGGER_OVERWRITE
#endif

// Soft breakpoint timeout in seconds
#ifndef DBG_SOFT_BKP_TIMEOUT
#define DBG_SOFT_BKP_TIMEOUT 120
#endif

// Turn off execution profiler
#ifdef LIB_PERF_PROFILING
#undef LIB_PERF_PROFILING
#endif

//
/////////////////////////////////////////////////////////////////////////////////////////////////////





/////////////////////////////////////////////////////////////////////////////////////////////////////
//
// Internal helper definitions
//

// Execution delay
#ifdef DBG_TRACE_EXEC_DELAY_MS
#define DBG_TRACE_EXEC_DELAY { ulong dbg_trace_delay_end = (GetTickCount64() + DBG_TRACE_EXEC_DELAY_MS); while((!_DEBUG) && (!_StopFlag) && (dbg_trace_delay_end > GetTickCount64())) { Sleep(100); }; }
#define DBG_STR_TRACE_EXEC_DELAY_INFO StringFormat("Debug trace execution delay: %i ms", DBG_TRACE_EXEC_DELAY_MS)
#else
#define DBG_TRACE_EXEC_DELAY
#define DBG_STR_TRACE_EXEC_DELAY_INFO ""
#endif

// Soft breakpoint timeout command
#define DBG_SLEEP_SECONDS(x) { ulong dbg_sleep_seconds_delay = (GetTickCount64() + (x * 1000)); while((!_DEBUG) && (!_StopFlag) && (dbg_sleep_seconds_delay > GetTickCount64())) { Sleep(100); }; }

// Debugger state string
#define DBG_DEBUGGER_FLAG_STATE "[DEBUG]"

// Debugger runtime flag state
#define DBG_DEBUGGER_RUNTIME_STATE StringFormat("IS_DEBUG_MODE = %s\r\n     %s", (IS_DEBUG_MODE) ? "true" : "false", DBG_STR_TRACE_EXEC_DELAY_INFO)

// EX5 and terminal info string
#define DBG_STR_EX5_FILEINFO StringFormat("MQL5 compiler build: %i %s\r\n  compiled@%s\r\n  => %s\r\n     %s\r\n     %s", __MQL5BUILD__, DBG_DEBUGGER_FLAG_STATE, TimeToString(__DATETIME__), DBG_DEBUGGER_FLAG_STATE, DBG_DEBUGGER_FLAG, DBG_DEBUGGER_RUNTIME_STATE)

//
/////////////////////////////////////////////////////////////////////////////////////////////////////





/////////////////////////////////////////////////////////////////////////////////////////////////////
//
// Debugger helpers
//
    // Overloaded var_out function
    #ifndef LIB_DEBUG_NAMESPACE
    namespace dbg_lib
    {
        static int trace_call_depth = NULL; 
        
        const string var_out(const string name, const string val)   { return(StringFormat("[str]  %-60s = %s", name, val)); }
        const string var_out(const string name, const char val)     { return(StringFormat("[chr]  %-60s = %-25hd (0x%hX)", name, val, (uchar)val)); }
        const string var_out(const string name, const short val)    { return(StringFormat("[shrt] %-60s = %-25hi (0x%hX)", name, val, val)); }
        const string var_out(const string name, const int val)      { return(StringFormat("[int]  %-60s = %-25i (0x%X)", name, val, val)); }
        const string var_out(const string name, const long val)     { return(StringFormat("[long] %-60s = %-25lli (0x%llX)", name, val, val)); }
        const string var_out(const string name, const uchar val)    { return(StringFormat("[uchr] %-60s = %-25hu (0x%hX)", name, val, val)); }
        const string var_out(const string name, const ushort val)   { return(StringFormat("[usht] %-60s = %-25hu (0x%hX)", name, val, val)); }
        const string var_out(const string name, const uint val)     { return(StringFormat("[uint] %-60s = %-25u (0x%X)", name, val, val)); }
        const string var_out(const string name, const ulong val)    { return(StringFormat("[ul]   %-60s = %-25llu (0x%llX)", name, val, val)); }
        const string var_out(const string name, const float val)    { return(StringFormat("[flt]  %-60s = %" + (StringFormat("%i", -30 - DBG_FLOAT_OUTPUT_DIGITS)) + "s (%#a)", name, StringFormat("%.*f",DBG_FLOAT_OUTPUT_DIGITS, val), val)); }
        const string var_out(const string name, const double val)   { return(StringFormat("[dbl]  %-60s = %" + (StringFormat("%i", -30 - DBG_DOUBLE_OUTPUT_DIGITS)) + "s (%#a)", name, StringFormat("%.*f", DBG_DOUBLE_OUTPUT_DIGITS, val), val)); }
        const string var_out(const string name, const bool val)     { return(StringFormat("[bool] %-60s = %s", name, (val) ? "true" : "false")); }
        const string var_out(const string name, const datetime val) { return(StringFormat("[dtm]  %-60s = %s", name, (val == NULL) ? "NULL" : ((val < NULL) ? StringFormat("[NEGATIVE VALUE] %lli", val) : ((StringLen(TimeToString(val, TIME_DATE | TIME_SECONDS)) <= NULL) ? StringFormat("[NON DATE VALUE] %llu", val) : TimeToString(val, TIME_DATE | TIME_SECONDS)) ))); }
        template <typename T> 
        const string var_out(const string name, const T val)        { return(StringFormat("[enum] %-60s = %s", name, EnumToString((T)val))); }
        
        void uninit_text()
        {
            switch(_UninitReason)
            {
                case REASON_PROGRAM:        printf(DBG_OUTPUT_PREFIX + DBG_OUTPUT_PREFIX + DBG_OUTPUT_FORMAT("", ((StringFormat("%i = REASON_PROGRAM: %s",      (int)_UninitReason, "Expert Advisor terminated its operation by calling the ExpertRemove() function."))))); return;
                case REASON_REMOVE:         printf(DBG_OUTPUT_PREFIX + DBG_OUTPUT_PREFIX + DBG_OUTPUT_FORMAT("", ((StringFormat("%i = REASON_REMOVE: %s",       (int)_UninitReason, "Program has been deleted from the chart."))))); return;
                case REASON_RECOMPILE:      printf(DBG_OUTPUT_PREFIX + DBG_OUTPUT_PREFIX + DBG_OUTPUT_FORMAT("", ((StringFormat("%i = REASON_RECOMPILE: %s",    (int)_UninitReason, "Program has been recompiled."))))); return;
                case REASON_CHARTCHANGE:    printf(DBG_OUTPUT_PREFIX + DBG_OUTPUT_PREFIX + DBG_OUTPUT_FORMAT("", ((StringFormat("%i = REASON_CHARTCHANGE: %s",  (int)_UninitReason, "Symbol or chart period has been changed."))))); return;
                case REASON_CHARTCLOSE:     printf(DBG_OUTPUT_PREFIX + DBG_OUTPUT_PREFIX + DBG_OUTPUT_FORMAT("", ((StringFormat("%i = REASON_CHARTCLOSE: %s",   (int)_UninitReason, "Chart has been closed."))))); return;
                case REASON_PARAMETERS:     printf(DBG_OUTPUT_PREFIX + DBG_OUTPUT_PREFIX + DBG_OUTPUT_FORMAT("", ((StringFormat("%i = REASON_PARAMETERS: %s",   (int)_UninitReason, "Input parameters have been changed by a user."))))); return;
                case REASON_ACCOUNT:        printf(DBG_OUTPUT_PREFIX + DBG_OUTPUT_PREFIX + DBG_OUTPUT_FORMAT("", ((StringFormat("%i = REASON_ACCOUNT: %s",      (int)_UninitReason, "Another account has been activated or reconnection to the trade server has occurred due to changes in the account settings."))))); return;
                case REASON_TEMPLATE:       printf(DBG_OUTPUT_PREFIX + DBG_OUTPUT_PREFIX + DBG_OUTPUT_FORMAT("", ((StringFormat("%i = REASON_TEMPLATE: %s",     (int)_UninitReason, "A new template has been applied."))))); return;
                case REASON_INITFAILED:     printf(DBG_OUTPUT_PREFIX + DBG_OUTPUT_PREFIX + DBG_OUTPUT_FORMAT("", ((StringFormat("%i = REASON_INITFAILED: %s",   (int)_UninitReason, "This value means that OnInit() handler has returned a nonzero value."))))); return;
                case REASON_CLOSE:          printf(DBG_OUTPUT_PREFIX + DBG_OUTPUT_PREFIX + DBG_OUTPUT_FORMAT("", ((StringFormat("%i = REASON_CLOSE: %s",        (int)_UninitReason, "Terminal has been closed."))))); return;
            }
            printf(DBG_OUTPUT_PREFIX + DBG_OUTPUT_PREFIX + DBG_OUTPUT_FORMAT("", ((StringFormat("Unknown reason: %i", _UninitReason)))));
            return;
        }
    };
    #endif


    // Debug to string functions
    #define DBG_STR(x)                      StringFormat(DBG_OUTPUT_FORMAT("", x))
    #define DBG_STR_PERSIST(x)              DBG_STR(x)
    #define DBG_STR_VAR(x)                  (dbg_lib::var_out(#x, x))

    // Debug std out
    #define DBG_MSG(x)                      printf("%*s%s", dbg_lib::trace_call_depth * 2, "", StringFormat(DBG_OUTPUT_PREFIX + DBG_OUTPUT_PREFIX + DBG_OUTPUT_FORMAT("", (x))))
    #define DBG_MSG_PERSIST(x)              DBG_MSG(x)
    #define DBG_MSG_VAR(x)                  DBG_MSG(DBG_STR_VAR((x)))

    // Code tracing helpers
    #define DBG_MSG_TRACE_BEGIN             { dbg_lib::trace_call_depth++; printf("\r\n%*s%s", dbg_lib::trace_call_depth * 2, "", StringFormat(DBG_OUTPUT_PREFIX + DBG_OUTPUT_FORMAT(DBG_TRACE_BEGIN_MSG_TXT, StringFormat("  v-v-v-v-v [ Call depth: %-3i] v-v-v-v-v [ %s()  BEGIN ] v-v-v-v-v-v-v-v-v-v-v-v-v-v-v-v-v", dbg_lib::trace_call_depth, __FUNCTION__)))); }
    #define DBG_MSG_TRACE_END               { printf("%*s%s", dbg_lib::trace_call_depth * 2, "", StringFormat(DBG_OUTPUT_PREFIX + DBG_OUTPUT_FORMAT(DBG_TRACE_END_MSG_TXT, StringFormat("    ^^^^^^^^^ [ Call depth: %-3i] ^^^^^^^^^ [ %s()  END ] ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^", dbg_lib::trace_call_depth, __FUNCTION__)))); printf("%s", " "); dbg_lib::trace_call_depth--; DBG_TRACE_EXEC_DELAY; }
    #define DBG_MSG_TRACE_RETURN            DBG_MSG("=> return(void)"); DBG_MSG_TRACE_END return
    #define DBG_MSG_TRACE_RETURN_VAR(x)     DBG_MSG(StringFormat("=> return(%s %s", typename((x)), DBG_STR_VAR(x)) + ");"); DBG_MSG_TRACE_END return(x)
    #define DBG_MSG_NOTRACE_RETURN          return
    #define DBG_MSG_NOTRACE_RETURN_VAR(x)   return(x)

    // Debug comments
    #define DBG_MSG_UNINIT_RESOLVER         if(_StopFlag) { DBG_MSG_VAR(_StopFlag); } dbg_lib::uninit_text();
    #define DBG_STR_COMMENT(cmnt)           cmnt

    // Class and object polymorphism
    #define DBG_STR_OBJ_RTTI                DBG_STR(StringFormat("%s %d", typename(this), &this))
    #define DBG_MSG_OBJ_RTTI                DBG_MSG(StringFormat("%s %d", typename(this), &this))

//
/////////////////////////////////////////////////////////////////////////////////////////////////////





/////////////////////////////////////////////////////////////////////////////////////////////////////
//
// Vardump array
//

    // Vardump pause execution
    #ifdef DBG_VARDUMP_PAUSE
    #undef DBG_VARDUMP_PAUSE_COMMAND
//    #define DBG_VARDUMP_PAUSE_COMMAND(x) DBG_MSG(StringFormat("Vardump: %s %s[%i] done. Execution delayed for %i seconds.", typename((x)), #x, ArraySize(x), DBG_VARDUMP_PAUSE)); DBG_SLEEP_SECONDS(DBG_VARDUMP_PAUSE)
    #else
    #undef DBG_VARDUMP_PAUSE_COMMAND
    #define DBG_VARDUMP_PAUSE_COMMAND(x) DBG_MSG(StringFormat("Vardump: %s %s[%i] done.", typename((x)), #x, ArraySize(x)));
    #endif

    // Dump function
    #ifndef LIB_DEBUG_NAMESPACE
    namespace dbg_lib
    {
        template <typename T> void VarDump(const T& _array_[], const string _var_name_, const string _file_, const string _func_, const int _line_)
        {
            string type = typename(_array_);
            printf(DBG_OUTPUT_STRING, "", _file_, _func_, _line_, StringFormat("Vardump array: %s %s[%i] type: %s", type, _var_name_, ArraySize(_array_), ((ArrayIsDynamic(_array_)) ? "dynamic" : "static")));
            if(type == "string")
            { for(int cnt = 0x00; cnt < ArraySize(_array_); cnt++) { printf("%s%s%s%s[%i] = %s ", DBG_OUTPUT_PREFIX, DBG_OUTPUT_PREFIX, DBG_OUTPUT_PREFIX, _var_name_, cnt, _array_[cnt]); } }
            else if(type == "double")
            { for(int cnt = 0x00; cnt < ArraySize(_array_); cnt++) { printf("%s%s%s%s[%i] = %.*f ", DBG_OUTPUT_PREFIX, DBG_OUTPUT_PREFIX, DBG_OUTPUT_PREFIX, _var_name_, cnt, DBG_DOUBLE_OUTPUT_DIGITS, _array_[cnt]); } }
            else if(type == "bool")
            { for(int cnt = 0x00; cnt < ArraySize(_array_); cnt++) { printf("%s%s%s%s[%i] = %s ", DBG_OUTPUT_PREFIX, DBG_OUTPUT_PREFIX, DBG_OUTPUT_PREFIX, _var_name_, cnt, ((((int)_array_[cnt]) != NULL) ? "true" : "false")); } }
            else if(type == "datetime")
            { for(int cnt = 0x00; cnt < ArraySize(_array_); cnt++) { printf("%s%s%s%s[%i] = %s ", DBG_OUTPUT_PREFIX, DBG_OUTPUT_PREFIX, DBG_OUTPUT_PREFIX, _var_name_, cnt, TimeToString((datetime)_array_[cnt])); } }
            else
            { for(int cnt = 0x00; cnt < ArraySize(_array_); cnt++) { printf("%s%s%s%s[%i] = %i ", DBG_OUTPUT_PREFIX, DBG_OUTPUT_PREFIX, DBG_OUTPUT_PREFIX, _var_name_, cnt, _array_[cnt]); } }
            if(!_DEBUG)
            { printf(DBG_OUTPUT_STRING, "", _file_, _func_, _line_, StringFormat("Vardump: %s %s[%i] done. Execution delayed for %i seconds.", type, _var_name_, ArraySize(_array_), DBG_VARDUMP_PAUSE)); DBG_SLEEP_SECONDS(DBG_VARDUMP_PAUSE); }
            else
            { printf(DBG_OUTPUT_STRING, "", _file_, _func_, _line_, StringFormat("Vardump: %s %s[%i] done.", type, _var_name_, ArraySize(_array_))); }
        return;
        }        
        template <typename T> const string VoidToString(const T value)
        {
            string type = typename(value);
            if(type == "string")
            { return(StringFormat("%s", value)); }
            else if(type == "double")
            { return(StringFormat("%.*f", DBG_DOUBLE_OUTPUT_DIGITS, value)); }
            else if(type == "bool")
            { return(StringFormat("%s", ((((int)value) != NULL) ? "true" : "false"))); }
            else if(type == "datetime")
            { return(StringFormat("%s", TimeToString((datetime)value))); }
            return(StringFormat("%i", value));
        }
        template <typename T, typename U> void VarDump(const T& _array1_[], const string _var_name1_, const U& _array2_[], const string _var_name2_, const string _file_, const string _func_, const int _line_)
        {
            printf(DBG_OUTPUT_STRING, "", _file_, _func_, _line_, StringFormat("Vardump list: %s %s[%i] type: %s; %s %s[%i] type: %s", typename(_array1_), _var_name1_, ArraySize(_array1_), ((ArrayIsDynamic(_array1_)) ? "dynamic" : "static"), typename(_array2_), _var_name2_, ArraySize(_array2_), ((ArrayIsDynamic(_array2_)) ? "dynamic" : "static")));
            for(int cnt = 0x00; cnt < MathMin(ArraySize(_array1_), ArraySize(_array2_)); cnt++) { printf("%s%s%s%s[%i], %30s[%i] = %-30s -> %s", DBG_OUTPUT_PREFIX, DBG_OUTPUT_PREFIX, DBG_OUTPUT_PREFIX, _var_name1_, cnt, _var_name2_, cnt, VoidToString(_array1_[cnt]), VoidToString(_array2_[cnt])); }
            if(!_DEBUG)
            { printf(DBG_OUTPUT_STRING, "", _file_, _func_, _line_, StringFormat("Vardump: %s %s[%i] -> %s %s[%i] done. Execution delayed for %i seconds.", typename(_array1_), _var_name1_, ArraySize(_array1_), typename(_array2_), _var_name2_, ArraySize(_array2_), DBG_VARDUMP_PAUSE)); DBG_SLEEP_SECONDS(DBG_VARDUMP_PAUSE); }
            else
            { printf(DBG_OUTPUT_STRING, "", _file_, _func_, _line_, StringFormat("Vardump: %s %s[%i] -> %s %s[%i] done.", typename(_array1_), _var_name1_, ArraySize(_array1_), typename(_array1_), _var_name1_, ArraySize(_array1_))); }
        return;
        }

        template <typename T> void VarDump(const T _value_, const string _var_name_, const string _file_, const string _func_, const int _line_)
        {
            string type = typename(_value_);
            printf(DBG_OUTPUT_STRING, "", _file_, _func_, _line_, StringFormat("Vardump value: %s %s", type, _var_name_));
            if(type == "string")
            { printf("%s%s%s%s = %s ", DBG_OUTPUT_PREFIX, DBG_OUTPUT_PREFIX, DBG_OUTPUT_PREFIX, _var_name_, _value_); }
            else if(type == "double")
            { printf("%s%s%s%s = %." + IntegerToString(DBG_DOUBLE_OUTPUT_DIGITS) + "f ", DBG_OUTPUT_PREFIX, DBG_OUTPUT_PREFIX, DBG_OUTPUT_PREFIX, _var_name_, _value_); }
            else if(type == "bool")
            { printf("%s%s%s%s = %s ", DBG_OUTPUT_PREFIX, DBG_OUTPUT_PREFIX, DBG_OUTPUT_PREFIX, _var_name_, ((_value_ != NULL) ? "true" : "false")); }
            else if(type == "datetime")
            { printf("%s%s%s%s = %s ", DBG_OUTPUT_PREFIX, DBG_OUTPUT_PREFIX, DBG_OUTPUT_PREFIX, _var_name_, TimeToString((datetime)_value_)); }
            else
            { printf("%s%s%s%s = %i ", DBG_OUTPUT_PREFIX, DBG_OUTPUT_PREFIX, DBG_OUTPUT_PREFIX, _var_name_, _value_); }
            if(!_DEBUG)
            { printf(DBG_OUTPUT_STRING, "", _file_, _func_, _line_, StringFormat("Vardump: %s %s done. Execution delayed for %i seconds.", type, _var_name_, DBG_VARDUMP_PAUSE)); DBG_SLEEP_SECONDS(DBG_VARDUMP_PAUSE); }
            else
            { printf(DBG_OUTPUT_STRING, "", _file_, _func_, _line_, StringFormat("Vardump: %s %s done.", type, _var_name_)); }
        return;
        }
    };
    #endif 
    
    // Var dump macro
    #define DBG_MSG_VARDUMP(x)      dbg_lib::VarDump(x, #x, __FILE__, __FUNCSIG__, __LINE__)
    #define DBG_MSG_LISTDUMP(x, y)  dbg_lib::VarDump(x, #x, y, #y, __FILE__, __FUNCSIG__, __LINE__)

//
/////////////////////////////////////////////////////////////////////////////////////////////////////





/////////////////////////////////////////////////////////////////////////////////////////////////////
//
// Assert support
//

    // Generic assert definition
    #define DBG_GENERIC_ASSERT(condition, message) { if(!(condition)) { string msg = StringFormat(DBG_OUTPUT_FORMAT(#condition, message));

    // Call modal dialog and div/zero!
    #define DBG_ASSERT(condition, message)                      	DBG_GENERIC_ASSERT(condition, message) Alert(DBG_ASSERT_MSG_TXT + " " + msg); DBG_CRASH_CODE; } }

    // Print out information
    #define DBG_ASSERT_LOG(condition, message)                  	DBG_GENERIC_ASSERT(condition, message) printf(DBG_OUTPUT_PREFIX + DBG_OUTPUT_PREFIX + "%s %s", DBG_ASSERT_MSG_TXT, msg); } }

    // Print info and abort execution by return;
    #define DBG_ASSERT_RETURN(condition, message)          			DBG_GENERIC_ASSERT(condition, message) printf(DBG_OUTPUT_PREFIX + DBG_OUTPUT_PREFIX + "%s %s", DBG_ASSERT_MSG_TXT, msg); return; } }

    // Print info and abort execution by return;
    #define DBG_ASSERT_RETURN_VAR(condition, message, return_value) DBG_GENERIC_ASSERT(condition, message) printf(DBG_OUTPUT_PREFIX + DBG_OUTPUT_PREFIX + "%s %s", DBG_ASSERT_MSG_TXT, msg); return(return_value); } }

//
/////////////////////////////////////////////////////////////////////////////////////////////////////





/////////////////////////////////////////////////////////////////////////////////////////////////////
//
// Software inherited brakepoints
//

    // Stop function
    #ifndef LIB_DEBUG_NAMESPACE
    #define LIB_DEBUG_NAMESPACE
    namespace dbg_lib
    {
        const bool BreakPoint(const int _line_, const string _file_, const long _timestamp_ = NULL)
        {
            #ifndef DBG_DEBUGGER_OVERWRITE
            if(IS_DEBUG_MODE)
            { printf("##### Breakpoint set at line: %i in file: %s.", _line_, _file_); return(true); }
            else
            #endif
            if(_timestamp_ <= ((TimeCurrent() - (TimeCurrent() % PeriodSeconds())) * (_timestamp_ > NULL)))
            { printf("##### Breakpoint set at line: %i in file: %s, waiting %i seconds.", _line_, _file_, DBG_SOFT_BKP_TIMEOUT); return(true); }
            return(false);
        }
        const bool ExecTime(const uint _seconds_)
        {
            static ulong exec_time_start = GetMicrosecondCount();
            exec_time_start = (exec_time_start * (ulong)((_seconds_ > 0x00) & 0x01)) + (GetMicrosecondCount() * (ulong)((_seconds_ == NULL) & 0x01));
            return(((long)exec_time_start) < ((long)(GetMicrosecondCount() - (_seconds_ * 1000000))));
        }
    };
    #endif

    // Debugging macros
    #define DBG_SOFT_BREAKPOINT                 { dbg_lib::BreakPoint(__LINE__, __FILE__); if(IS_DEBUG_MODE) { DebugBreak(); } else { DBG_SLEEP_SECONDS(DBG_SOFT_BKP_TIMEOUT); } } 
    #define DBG_SOFT_BREAKPOINT_TS(timestamp)   { dbg_lib::BreakPoint(__LINE__, __FILE__, StringToTime(timestamp)); DBG_SOFT_BREAKPOINT }
    #define DBG_SOFT_BREAKPOINT_CONDITION(x)    { if((x) && (dbg_lib::BreakPoint(__LINE__, __FILE__, TimeCurrent() + DBG_SOFT_BKP_TIMEOUT))) { DBG_SOFT_BREAKPOINT; } }
    #define DBG_SOFT_BREAKPOINT_EXEC_TIME(x)    { DBG_SOFT_BREAKPOINT_CONDITION(dbg_lib::ExecTime(x)); dbg_lib::ExecTime(NULL); }

//
/////////////////////////////////////////////////////////////////////////////////////////////////////





/////////////////////////////////////////////////////////////////////////////////////////////////////
//
// Loop tracing
//

    // Init loop trace
    #define DBG_TRACE_LOOP_BEGIN_ID(x)  int _dbg_loop_start##x = 0x00; int _dbg_loop_finish##x = 0x00; ulong _dbg_loop_runtime##x = GetMicrosecondCount();
    #define DBG_TRACE_LOOP_BEGIN        DBG_TRACE_LOOP_BEGIN_ID(__FUNCTION__)

    // Count loop heads
    #define DBG_TRACE_LOOP_START_ID(x)  _dbg_loop_start##x++;
    #define DBG_TRACE_LOOP_START        DBG_TRACE_LOOP_START_ID(__FUNCTION__)

    // Count loop footers
    #define DBG_TRACE_LOOP_FINISH_ID(x) _dbg_loop_finish##x++;
    #define DBG_TRACE_LOOP_FINISH       DBG_TRACE_LOOP_FINISH_ID(__FUNCTION__)

    // Print loop stats
    #define DBG_TRACE_LOOP_END_ID(x)    DBG_MSG(StringFormat("Iterations begin: %i; Iterations end: %i; Runtime: %i microseconds.", _dbg_loop_start##x, _dbg_loop_finish##x, (GetMicrosecondCount() - _dbg_loop_runtime##x)));
    #define DBG_TRACE_LOOP_END          DBG_TRACE_LOOP_END_ID(__FUNCTION__)

//
/////////////////////////////////////////////////////////////////////////////////////////////////////





/////////////////////////////////////////////////////////////////////////////////////////////////////
//
// Runtime code setup
//
#else
    // Define runtime mode
    
        #define LIB_DEBUG_MQH_RUNTIME_MODE
        
        
    // Clear any existing definitions

        #ifdef DBG_CODE_LOCATION
        #undef DBG_CODE_LOCATION
        #endif
        #ifdef DBG_DEBUGGER_FLAG_STATE
        #undef DBG_DEBUGGER_FLAG_STATE
        #endif
        #ifdef DBG_STR_EX5_FILEINFO
        #undef DBG_STR_EX5_FILEINFO
        #endif
        #ifdef DBG_STR
        #undef DBG_STR
        #endif
        #ifdef DBG_STR_PERSIST
        #undef DBG_STR_PERSIST
        #endif
        #ifdef DBG_STR_VAR
        #undef DBG_STR_VAR
        #endif
        #ifdef DBG_MSG
        #undef DBG_MSG
        #endif
        #ifdef DBG_MSG_PERSIST
        #undef DBG_MSG_PERSIST
        #endif
        #ifdef DBG_MSG_VAR
        #undef DBG_MSG_VAR
        #endif
        #ifdef DBG_MSG_VARDUMP
        #undef DBG_MSG_VARDUMP
        #endif
        #ifdef DBG_MSG_LISTDUMP
        #undef DBG_MSG_LISTDUMP
        #endif
        #ifdef DBG_MSG_TRACE_BEGIN
        #undef DBG_MSG_TRACE_BEGIN
        #endif
        #ifdef DBG_MSG_TRACE_END
        #undef DBG_MSG_TRACE_END
        #endif
        #ifdef DBG_MSG_TRACE_RETURN
        #undef DBG_MSG_TRACE_RETURN
        #endif
        #ifdef DBG_MSG_TRACE_RETURN_VAR
        #undef DBG_MSG_TRACE_RETURN_VAR
        #endif
        #ifdef DBG_MSG_NOTRACE_RETURN
        #undef DBG_MSG_NOTRACE_RETURN
        #endif
        #ifdef DBG_MSG_NOTRACE_RETURN_VAR
        #undef DBG_MSG_NOTRACE_RETURN_VAR
        #endif
        #ifdef DBG_MSG_UNINIT_RESOLVER
        #undef DBG_MSG_UNINIT_RESOLVER
        #endif
        #ifdef DBG_STR_COMMENT
        #undef DBG_STR_COMMENT
        #endif
        #ifdef DBG_ASSERT
        #undef DBG_ASSERT
        #endif
        #ifdef DBG_ASSERT_LOG
        #undef DBG_ASSERT_LOG
        #endif
        #ifdef DBG_ASSERT_RETURN
        #undef DBG_ASSERT_RETURN
        #endif
        #ifdef DBG_ASSERT_RETURN_VAR
        #undef DBG_ASSERT_RETURN_VAR
        #endif
        #ifdef DBG_SLEEP_SECONDS
        #undef DBG_SLEEP_SECONDS
        #endif
        #ifdef DBG_SOFT_BREAKPOINT
        #undef DBG_SOFT_BREAKPOINT
        #endif
        #ifdef DBG_SOFT_BREAKPOINT_TS
        #undef DBG_SOFT_BREAKPOINT_TS
        #endif
        #ifdef DBG_SOFT_BREAKPOINT_CONDITION
        #undef DBG_SOFT_BREAKPOINT_CONDITION
        #endif
        #ifdef DBG_SOFT_BREAKPOINT_EXEC_TIME
        #undef DBG_SOFT_BREAKPOINT_EXEC_TIME
        #endif
        #ifdef DBG_TRACE_LOOP_BEGIN
        #undef DBG_TRACE_LOOP_BEGIN
        #endif
        #ifdef DBG_TRACE_LOOP_START
        #undef DBG_TRACE_LOOP_START
        #endif
        #ifdef DBG_TRACE_LOOP_FINISH
        #undef DBG_TRACE_LOOP_FINISH
        #endif
        #ifdef DBG_TRACE_LOOP_END
        #undef DBG_TRACE_LOOP_END
        #endif
        #ifdef DBG_TRACE_LOOP_BEGIN_ID
        #undef DBG_TRACE_LOOP_BEGIN_ID
        #endif
        #ifdef DBG_TRACE_LOOP_START_ID
        #undef DBG_TRACE_LOOP_START_ID
        #endif
        #ifdef DBG_TRACE_LOOP_FINISH_ID
        #undef DBG_TRACE_LOOP_FINISH_ID
        #endif
        #ifdef DBG_TRACE_LOOP_END_ID
        #undef DBG_TRACE_LOOP_END_ID
        #endif


    // EX5 File info string

        #define DBG_DEBUGGER_FLAG_STATE "[RELEASE]"
        #define DBG_STR_EX5_FILEINFO StringFormat("MQL5 build: %i %s", __MQL5BUILD__, DBG_DEBUGGER_FLAG_STATE)


    // Remove debug output and comments

        // Location string
        #define DBG_CODE_LOCATION(x)
        
        // Debug to string functions
        #define DBG_STR(x)
        #define DBG_STR_PERSIST(x) x
        #define DBG_STR_VAR(x)

        // Debug std out
        #define DBG_MSG(x)
        #define DBG_MSG_PERSIST(x) printf(x)
        #define DBG_MSG_VAR(x)
        #define DBG_MSG_VARDUMP(x)
        #define DBG_MSG_LISTDUMP(x, y)

        // Code tracing helpers
        #define DBG_MSG_TRACE_BEGIN
        #define DBG_MSG_TRACE_END
        #define DBG_MSG_TRACE_RETURN            PERF_COUNTER_END return
        #define DBG_MSG_TRACE_RETURN_VAR(x)     PERF_COUNTER_END return(x)
        #define DBG_MSG_NOTRACE_RETURN          PERF_COUNTER_END return
        #define DBG_MSG_NOTRACE_RETURN_VAR(x)   PERF_COUNTER_END return(x)

        // Debug comments
        #define DBG_MSG_UNINIT_RESOLVER
        #define DBG_STR_COMMENT(x) NULL


    // Remove asserts

        #define DBG_ASSERT(condition, message)
        #define DBG_ASSERT_LOG(condition, message)
        #define DBG_ASSERT_RETURN_VOID(condition, message)
        #define DBG_ASSERT_RETURN(condition, message, return_value)


    // Remove soft break points

        #define DBG_SLEEP_SECONDS(x)
        #define DBG_SOFT_BREAKPOINT
        #define DBG_SOFT_BREAKPOINT_TS(x)
        #define DBG_SOFT_BREAKPOINT_CONDITION(x)
        #define DBG_SOFT_BREAKPOINT_EXEC_TIME(x)


    // Loop tracing

        #define DBG_TRACE_LOOP_BEGIN
        #define DBG_TRACE_LOOP_START
        #define DBG_TRACE_LOOP_FINISH
        #define DBG_TRACE_LOOP_END
        #define DBG_TRACE_LOOP_BEGIN_ID(x)
        #define DBG_TRACE_LOOP_START_ID(x)
        #define DBG_TRACE_LOOP_FINISH_ID(x)
        #define DBG_TRACE_LOOP_END_ID(x)

//
/////////////////////////////////////////////////////////////////////////////////////////////////////





/////////////////////////////////////////////////////////////////////////////////////////////////////
//
// Performance profiling support
//
#ifdef LIB_PERF_PROFILING

    // Internal global counters
    #ifndef LIB_PERF_NAMESPACE
    #define LIB_PERF_NAMESPACE
    namespace lib_perf { static ulong perf_counters = 0x00; static ulong perf_array[]; static ulong perf_min[]; static ulong perf_max[]; }
    #endif

    // Performance output format string
    #define PERF_OUTPUT_FORMAT(x)       printf("%s() Runtime: %i microseconds; average(%i), min(%i), max(%i); total calls: %i", __FUNCTION__, _perf_runtime_##x, (lib_perf::perf_array[(int)_perf_id_##x] / _perf_calls_##x), lib_perf::perf_min[(int)_perf_id_##x], lib_perf::perf_max[(int)_perf_id_##x], _perf_calls_##x);

    // Performance arrays
    #define PERF_SET_ARRAYS             ((ulong)((ArrayResize(lib_perf::perf_array, (int)lib_perf::perf_counters) > NULL) && (ArrayResize(lib_perf::perf_min, (int)lib_perf::perf_counters) > NULL) && ((lib_perf::perf_min[(int)lib_perf::perf_counters - 1] = ((ulong)0x01 << 60)) > 0x00) && (ArrayResize(lib_perf::perf_max, (int)lib_perf::perf_counters) > 0x00)))
    #define PERF_UPDATE_COUNTERS(x)     lib_perf::perf_array[(int)_perf_id_##x] += _perf_runtime_##x; lib_perf::perf_min[(int)_perf_id_##x] = (((lib_perf::perf_min[(int)_perf_id_##x] > _perf_runtime_##x) || (lib_perf::perf_min[(int)_perf_id_##x] == 0x00)) ? _perf_runtime_##x : lib_perf::perf_min[(int)_perf_id_##x]); lib_perf::perf_max[(int)_perf_id_##x] = ((lib_perf::perf_max[(int)_perf_id_##x] < _perf_runtime_##x) ? _perf_runtime_##x : lib_perf::perf_max[(int)_perf_id_##x]);

    // Counter init
    #define PERF_COUNTER_DEFINE_ID(x)   const static ulong _perf_id_##x = lib_perf::perf_counters++; static ulong _perf_start_##x = PERF_SET_ARRAYS; static ulong _perf_calls_##x = (lib_perf::perf_array[(int)_perf_id_##x] = 0x00); _perf_calls_##x++;

    // Counter start
    #define PERF_COUNTER_SET_ID(x)      _perf_start_##x = GetMicrosecondCount();

    // Display performance counter
    #define PERF_COUNTER_CLOSE_ID(x)    { const ulong _perf_runtime_##x = (GetMicrosecondCount() - _perf_start_##x); PERF_UPDATE_COUNTERS(x); PERF_OUTPUT_FORMAT(x) }

    // Predefined function global performance counter
    #define PERF_COUNTER_BEGIN          PERF_COUNTER_DEFINE_ID(__FUNCTION__) PERF_COUNTER_SET_ID(__FUNCTION__)
    #define PERF_COUNTER_END            PERF_COUNTER_CLOSE_ID(__FUNCTION__)


#endif
#endif
//
/////////////////////////////////////////////////////////////////////////////////////////////////////





/////////////////////////////////////////////////////////////////////////////////////////////////////
//
// Remove performance profiling support
//
#ifndef LIB_PERF_PROFILING

    // Remove all performance collector code
    #ifndef PERF_COUNTER_DEFINE_ID
    #define PERF_COUNTER_DEFINE_ID(x)
    #endif

    #ifndef PERF_COUNTER_SET_ID
    #define PERF_COUNTER_SET_ID(x)
    #endif

    #ifndef PERF_COUNTER_CLOSE_ID
    #define PERF_COUNTER_CLOSE_ID(x)
    #endif

    #ifndef PERF_COUNTER_BEGIN
    #define PERF_COUNTER_BEGIN
    #endif

    #ifndef PERF_COUNTER_END
    #define PERF_COUNTER_END
    #endif


#endif
//
/////////////////////////////////////////////////////////////////////////////////////////////////////



//
// END Debugging support
//*********************************************************************************************************************************************************/
#endif // LIB_DEBUG_MQH_INCLUDED
